/*
 * Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
 *
 * This file is part of NRAO Workspaces.
 *
 * Workspaces is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Workspaces is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
 */

// Package db This file contains the methods and structs to get a connection to the archive
// database
package db

import (
	"database/sql"
	"fmt"
	_ "github.com/lib/pq"
	"gitlab.nrao.edu/ssa/gocapo/capo/config"
	capohelpers "gitlab.nrao.edu/ssa/gocapo/helpers"
	"strings"
)

// dbInfo Contains all necessary data to connect to the database
type dbInfo struct {
	host     string
	port     int
	user     string
	password string
	dbname   string
	isSsl    bool
}

// getDbLoginFromProperties Retrieve the archive database login details from a CAPO profile
//
// @param capoProfile a string with the CAPO profile to get the details from
// @param capoPath a string with the path to the CAPO profile
func getDbLoginFromProperties(capoProfile string) (dbInfo, error) {
	var connectionInfo dbInfo
	prop, err := config.InitConfig(capoProfile, capohelpers.DefaultCapoPath)
	if err != nil {
		return connectionInfo, err
	}

	// Prefix in the CAPO files for the archive database details
	prefix := "metadataDatabase"
	settings := prop.SettingsForPrefix(prefix)

	connectionInfo.user = settings.GetString("jdbcUsername")
	connectionInfo.password = settings.GetString("jdbcPassword")
	url := settings.GetString("jdbcUrl")
	url = strings.TrimPrefix(url, "jdbc:postgresql://")
	splitUrl := strings.Split(url, "/")
	connectionInfo.host = splitUrl[0]

	// Remove extraneous port on hostname if it exists
	connectionInfo.host = strings.Split(connectionInfo.host, ":")[0]

	connectionInfo.dbname = splitUrl[1]
	connectionInfo.port = 5432

	// Currently the archive database doesn't use SSL
	connectionInfo.isSsl = false

	return connectionInfo, err
}

// GetConnection Establish a connection to the database and return it
//
// @param capoProfile a string with the CAPO profile to get the database from
// @param capoPath a string with the path to the CAPO profile
func GetConnection(capoProfile string) (*sql.DB, error) {
	var db *sql.DB

	dbInfo, err := getDbLoginFromProperties(capoProfile)
	if err != nil {
		return db, err
	}

	sslMode := "disable"
	if dbInfo.isSsl {
		sslMode = "require"
	}

	connInfo := fmt.Sprintf(
		"host=%s port=%d user=%s password=%s dbname=%s sslmode=%s",
		dbInfo.host,
		dbInfo.port,
		dbInfo.user,
		dbInfo.password,
		dbInfo.dbname,
		sslMode,
	)

	db, err = sql.Open("postgres", connInfo)
	if err != nil {
		return db, err
	}

	err = db.Ping()
	if err != nil {
		return db, err
	}

	return db, err
}
