#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" This is the workflow service. """
import itertools

# pylint: disable=E0401, E0402, W1203

import json
import logging
import os
import pathlib
import shutil
import stat
import subprocess
from pathlib import Path
from tempfile import mkdtemp
from typing import Dict, List, Union

import requests
import transaction
from messaging.messenger import MessageSender
from messaging.router import Router, on_message
from pycapo import CapoConfig
from requests import Response

from workspaces.system.schema import AbstractFile
from workspaces.workflow.enum import WorkflowRequestState
from workspaces.workflow.message_architect import (
    ArchiveMessageArchitect,
    WorkflowMessageArchitect,
)
from workspaces.workflow.schema import Workflow, WorkflowRequest, WorkflowRequestFile
from workspaces.workflow.services.interfaces import WorkflowInfoIF, WorkflowServiceIF
from workspaces.workflow.services.monitor import WorkflowMonitor
from workspaces.workflow.services.recovery import MonitorRecover
from workspaces.workflow.services.remote_processing_service import CapoInjector

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class WorkflowServiceRESTClient(WorkflowServiceIF):
    """This is the workflow service REST client."""

    def __init__(self):
        self.url = CapoConfig().settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl

    def execute(self, request: WorkflowRequest):
        """
        Execute a workflow request.

        :param request: the workflow request to be executed
        :return: the response as JSON
        """
        response = requests.post(
            f"{self.url}/workflows/{request.workflow_name}/requests/{request.workflow_request_id}/submit"
        )
        logger.info(
            "Got result %s with type %s and body %s",
            response,
            type(response),
            response.content,
        )
        return response.json()

    def attach_file_to_request(self, request: WorkflowRequest, filename: str, content: bytes) -> Response:
        """
        Add a file to this workflow request.

        :param request: the workflow request
        :param filename: name of file to be added
        :param content: file contents
        :return: response to the PUT()
        """
        response = requests.put(
            f"{self.url}/workflows/{request.workflow_name}"
            f"/requests/{request.workflow_request_id}"
            f"/files/{filename}",
            data=content,
        )
        return response

    def retrieve_file_content(self, name: str, request_id: int, filename: str):
        """
        Retrieve the content of a workflow file
        :param name: the workflow name
        :param request_id: id of parent workflow request
        :param filename: file to find
        :return: dict containing file content
        """

        response = requests.get(f"{self.url}/workflows/{name}/requests/{request_id}/files/{filename}")
        return response.content.decode()

    def create_workflow_request(self, workflow: Union[str, Workflow], argument: Dict) -> WorkflowRequest:
        """
        Create a workflow request using the supplied arguments.

        :param workflow: workflow to be requested
        :param argument: arguments to the workflow request
        :return: the WorkflowRequest
        """

        # 1. Handle the type ambiguity with the workflow argument
        workflow_name = workflow if isinstance(workflow, str) else workflow.workflow_name

        # 2. Make the request
        result = requests.post(
            f"{self.url}/workflows/{workflow_name}/requests/create",
            json=argument,
        ).json()

        return WorkflowRequest.from_json(result)

    def announce_qa(self, workflow_request_id: int, msg_type: str, capability_version: int) -> WorkflowRequest:
        """
        Announce QA event

        :param workflow_request_id: id of workflow
        :param msg_type: pass/fail/ready
        :param capability_version: the version number
        :return:
        """
        argument = {"capabilityVersion": capability_version}

        result = requests.post(f"{self.url}/workflows/requests/{workflow_request_id}/qa/{msg_type}", json=argument)

        return WorkflowRequest.from_json(result.json())

    def ingest(self, request: WorkflowRequest):
        """
        Ingest results for previously run workflow into NGAS and archive
        :param request: completed workflow request to ingest
        :return:
        """
        requests.post(f"{self.url}/workflows/{request.workflow_name}/requests/{request.workflow_request_id}/ingest")


class WorkflowService(WorkflowServiceIF):
    """
    Executes workflows; should be a freestanding service.
    """

    # The next few things that need to happen here are in response to
    # WorkflowEvents that we receive from wf_monitor as the workflow
    # execution evolves. So we have to set up listening at some point
    # in this class
    def __init__(self, info: WorkflowInfoIF, message_handler: "WorkflowMessageHandler", recover: MonitorRecover):
        recover.reacquire_inflight_requests()
        self.info = info
        self.message_router = message_handler.message_router
        self.messenger = message_handler.messenger
        self.archive_messenger = message_handler.archive_messenger
        self.processing_settings = CapoConfig().settings("edu.nrao.workspaces.ProcessingSettings")
        self.external_controllers = info.all_external_controllers()

    def attach_file_to_request(self, request: WorkflowRequest, filename: str, content: bytes):
        """
        Add a file to a request; complain if the filename is bad.

        :param request: the request to which the file will be added
        :param filename: name of the file
        :param content: file contents

        :return:
        """
        forbidden = self._get_forbidden_templates_list(request)
        if filename not in forbidden:
            if (
                any(c.control_system == request.controller for c in self.external_controllers)
                and filename == "metadata.json"
            ):
                content_dict = json.loads(content.decode())
                content_dict["workflowName"] = request.workflow_name
                content_dict["data_location"] = request.argument["data_location"]
                content = json.dumps(content_dict, indent=4).encode()

            self.info.save_file(request=request, filename=filename, content=content)
        else:
            logger.info(f"{filename} is a protected file name.")

    def retrieve_file_content(self, request_id: int, filename: str):
        """
        Grab the content of this workflow request file.

        :param request_id:
        :param filename:
        :return:
        """
        wf_req = self.info.lookup_workflow_request(request_id)
        for file in wf_req.files:
            if file.filename == filename:
                return file.content.decode()

    def create_workflow_request(self, workflow, argument) -> WorkflowRequest:
        """
        Create a workflow request from the supplied arguments.

        :param workflow: the desired workflow
        :param argument: arguments to workflow request
        :return: the request that's been created
        """
        return self.info.create_workflow_request(workflow, argument)

    def send_carta_url_to_aat(self, workflow_request_id: int, carta_url: str):
        """
        Send AMQP message with CARTA url included to the AAT AMQP exchange

        :param workflow_request_id: ID of carta workflow that wants to send this message
        :param carta_url: JSON blob with CARTA URL
        """
        logger.info(f"SENDING CARTA MESSAGE to AAT Request Handler for request #{workflow_request_id}!")
        wf_request = self.info.lookup_workflow_request(workflow_request_id)
        routing_key = f"ws-workflow.carta-instance-ready.{workflow_request_id}"
        carta_url_msg = ArchiveMessageArchitect(
            routing_key=routing_key, request=wf_request, carta_url=carta_url
        ).compose_message("carta_ready")
        self.archive_messenger.send_message(**carta_url_msg)

    def send_carta_url(self, workflow_request_id: int, carta_url: str):
        """
        Send AMQP message with CARTA URL to Workspaces.

        :param workflow_request_id: ID of CARTA workflow that wants to send this message
        :param carta_url: JSON blob containing CARTA URL
        :return:
        """
        logger.info(f"SENDING CARTA MESSAGE to Workspaces for request #{workflow_request_id}!")
        wf_request = self.info.lookup_workflow_request(workflow_request_id)
        carta_url_msg = WorkflowMessageArchitect(request=wf_request, carta_url=carta_url).compose_message("carta_ready")
        self.messenger.send_message(**carta_url_msg)

    @staticmethod
    def _extract_sdm_id_from_request(request: WorkflowRequest):
        """
        Extract the SDM ID information from the information from the workflow
        arguments for this request.

        Note: Currently only used for std_calibration (since AAT/PPI only
        tracks calibration status for EBs, but added the imaging workflow
        for future work)

        :param request:
        :return:
        """

        argument_json = request.argument

        if request.workflow_name == "std_calibration":
            return argument_json["metadata"]["sdm_id"]
        elif request.workflow_name == "std_cms_imaging":
            return argument_json["sdmId"]
        else:
            return None

    def message_archive(self, identifier: Union[int, str], msg_type: str, additional_field: str = None):
        """
        Send a message to the AAT

        Note: removed the necessity to specify the SDM in arguments, since
        it is findable in the workflow arguments, and not readily available
        at the Action level of the CapabilityStateSystem

        :param identifier: either the workflow request id for the failed request, or the SDM ID needed by the archive
        :param msg_type: type of message to send
        :param additional_field: any additional information needed to send a message (usually a project code)
        :return:
        """
        msg_arch = None
        wf_request = None
        sdm_id = None
        project_code = None
        telescope = None

        if isinstance(identifier, int):
            # a request should exist, find it
            wf_request = self.info.lookup_workflow_request(identifier)
            if "parent_wf_request_id" in wf_request.argument:
                # this was a stand-alone workflow, look up the parent
                wf_request_usable = self.info.lookup_workflow_request(wf_request.argument["parent_wf_request_id"])
            else:
                wf_request_usable = wf_request

            sdm_id = self._extract_sdm_id_from_request(wf_request_usable)
            if additional_field is not None:
                project_code = additional_field
        else:
            # I'm initial ingestion, curation, or VLASS
            project_code = identifier
            # needed to determine indexer type in initial ingestion or curation
            if additional_field is not None:
                telescope = additional_field

        if msg_type == "do-not-calibrate":
            # this should only send from the WS QA interface
            logger.info(f"SENDING 'DO NOT CALIBRATE' message to AAT for request #{identifier}!")
            msg_arch = ArchiveMessageArchitect(routing_key="qa-script.fail", request=wf_request, sdm_id=sdm_id)
        elif msg_type == "cal-ingestion-complete":
            logger.info(f"SENDING 'Calibration Ingestion Complete' message to AAT for request #{identifier}!")
            msg_arch = ArchiveMessageArchitect(
                routing_key="ingestor.calibration",
                request=wf_request,
                sdm_id=sdm_id,
                project_code=project_code,
            )
        elif msg_type == "img-ingestion-complete":
            logger.info(f"Sending 'Image Ingestion Complete' indexing message to AAT for project {project_code}!")
            msg_arch = ArchiveMessageArchitect(routing_key="ingestor.images", request=None, project_code=project_code)
        elif msg_type == "obs-ingestion-complete":
            logger.info(f"Sending 'Observation Ingestion Complete' indexing message to AAT for project {project_code}!")
            key = "vlba" if telescope == "VLBA" or telescope == "GMVA" else "metadata"
            msg_arch = ArchiveMessageArchitect(
                routing_key=f"ingestion-complete.{key}", request=None, project_code=project_code, telescope=telescope
            )
        else:
            logger.info(f"Cannot determine archive message for type {msg_type}. Skipping....")

        msg = msg_arch.compose_message(msg_type.replace("-", "_"))
        self.archive_messenger.send_message(**msg)

    def execute(self, request: WorkflowRequest):
        """
        Execute a workflow per the supplied parameters.

        :param request: The Workflow Request to execute
        :return:
        """
        # look up workflow, returns workflow
        logger.info("Looking up workflow %s", request.workflow_name)
        definition = self.info.lookup_workflow_definition(request.workflow_name)

        # create a temporary directory if processing directory is not supplied,
        # needs to exist before template rendering
        temp_folder = self._make_temp_directory(request) if not request.results_dir else Path(request.results_dir)

        # Is this a remote execution workflow?
        injector = CapoInjector(self.info, request)
        remote = injector.is_remote_workflow()

        # if remote is true, create a capo subspace file in the request's directory
        if remote:
            logger.info("Creating Subspace File...")
            injector.subspace_to_directory()
            request.argument["profile"] = injector.profile

        # render templates to files, returns list of rendered files
        logger.info("Rendering files")

        # Aggrandize the argument with the workflow request ID and results directory
        request.argument["request_id"] = request.workflow_request_id
        request.argument["spool_dir"] = request.results_dir

        # Send results_dir to capability layer to store in the workflow metadata
        dir_info = WorkflowMessageArchitect(
            request=request, json_payload={"processing_dir": request.argument["spool_dir"]}
        ).compose_message("update_wf_metadata")

        # send the message
        self.messenger.send_message(**dir_info)

        # render all the templates
        if request.argument["need_project_metadata"] is True:
            templated_files = self._render_with_metadata(request, temp_folder, definition)
            # catch aat_wrest failure and abort workflow
            if isinstance(templated_files, WorkflowRequest):
                return request
        else:
            self.determine_multiple_productfetch(request)
            # will only be used if key is in the templates
            request.argument["ramInGb"] = self.processing_settings.ramInGb

            templated_files = definition.render_templates(request.argument)

        # Combine the templated with the request files.
        workflow_files = self._determine_usable_files(request, templated_files)

        # prepare files for condor execution
        self._check_for_old_condor_files(temp_folder)
        self._prepare_files_for_condor(workflow_files, temp_folder)
        self.info.save_request(request)

        # execute condor and retrieve log file
        [log_file, is_dag] = self._execute_prepared(Path(request.results_dir))

        # start reading the log file (background it)
        logger.info("Running wf_monitor.")
        WorkflowMonitor(log_file, request, is_dag).start()
        logger.info("Monitor running. Moving along...")

        return request

    def abort_running_workflow(self, request_id: int):
        """
        Given the ID for a running workflow request, attempt to abort the request's execution (using condor_rm)

        :param request_id: ID of running request
        """
        workflow_request = self.info.lookup_workflow_request(request_id)
        htcondor_job_id = workflow_request.htcondor_job_id
        if htcondor_job_id:
            subprocess.run(["condor_rm", f"{htcondor_job_id!s}"])
        else:
            logger.warning(
                f"Workflow request #{request_id} doesn't have an associated HTCondor job ID. Forcing fail transition..."
            )
            self.send_forced_fail(request_id)

        # Remove the results
        logger.info(f"Removing contents at {workflow_request.results_dir}")
        shutil.rmtree(workflow_request.results_dir)

    def _make_temp_directory(self, request: WorkflowRequest) -> Path:
        """
        Create tmp folder for workflow

        :return: path to tmp folder
        """
        vlapipe_id = 6000
        temp_folder_path = self.processing_settings.rootDirectory
        if not os.path.isdir(temp_folder_path):
            os.mkdir(temp_folder_path)
        temp_folder = Path(mkdtemp(dir=temp_folder_path))
        os.chown(temp_folder, vlapipe_id, vlapipe_id)
        os.chmod(temp_folder, 0o2770)
        logger.info("Settled on temp folder %s", temp_folder)
        # set new directory as result_dir
        request.results_dir = temp_folder.__str__()
        return temp_folder

    def determine_multiple_productfetch(self, wf_request: WorkflowRequest):
        """
        Are we fetching more than one product?

        :param wf_request: the workflow request
        :return:
        """
        if wf_request.workflow_name == "download" or (
            "need_data" in wf_request.argument and wf_request.argument["need_data"] is True
        ):
            spl_list = wf_request.argument["product_locator"].split(",")
            fetcher_string = " --product-locator"
            if "products" in wf_request.argument and (
                [True for product in wf_request.argument["products"] if product["product_type"] == "Image"]
            ):
                image_flag = "--image-ancillaries"
                wf_request.argument["fetch_args"] = (
                    image_flag + fetcher_string + fetcher_string.join(" " + spl for spl in spl_list)
                )
            else:
                wf_request.argument["fetch_args"] = fetcher_string + fetcher_string.join(" " + spl for spl in spl_list)

    def _get_wrester_args(self, wf_request: WorkflowRequest) -> List | WorkflowRequest:
        name = wf_request.workflow_name
        if "product_locator" in wf_request.argument:
            argument = wf_request.argument["product_locator"]
        elif "product_group_id" in wf_request.argument:
            argument = str(wf_request.argument["product_group_id"])
        else:
            argument = None
        argument2 = []

        def in_name(name_element: str) -> bool:
            """
            Check if a given string is contained in the workflow_name for a given workflow request
            Helper function to minimize code.

            :param name_element: string to search for
            :return: True if string in workflow_name
            """
            return name_element in wf_request.workflow_name

        if in_name("download"):
            wrest_type = "--product"
        elif in_name("calibration"):
            wrest_type = "-sc"
        elif in_name("restore"):
            wrest_type = "-r"
            argument2 = [wf_request.argument["cal_locator"]]
        elif in_name("imaging"):
            wrest_type = "-si"
            argument = wf_request.argument["sdmId"]
        elif in_name("carta"):
            wrest_type = "--carta"
        elif in_name("coarse"):
            wrest_type = "--cube"
            argument = wf_request.argument["precursor"]
        elif in_name("ingest"):
            wrest_type = "-aux"
            if "seci" in name:
                argument = wf_request.argument["cal_spl"]
            else:
                parent_req = self.info.lookup_workflow_request(int(wf_request.argument["parent_wf_request_id"]))
                eb = (
                    parent_req.argument["product_locator"]
                    if "product_locator" in parent_req.argument
                    else parent_req.argument["sdmId"]
                )
                argument = eb
        elif in_name("curator"):
            wrest_type = "--curator_products" if wf_request.argument["curator_type"].lower() == "full" else "--curator"
        else:
            logger.info(f"No wrester found for workflow {name}. Does it actually require metadata?")
            return wf_request

        return [wrest_type, argument, argument2]

    def _render_with_metadata(self, wf_request: WorkflowRequest, tempdir: Path, wf_definition: Workflow):
        name = wf_request.workflow_name
        arg_list = self._get_wrester_args(wf_request)

        if isinstance(arg_list, WorkflowRequest):
            # we failed to determine a wrester and received back the request object
            return arg_list
        else:
            logger.info(f" workflow {name} has wrest option: {arg_list[0]}")
            with subprocess.Popen(
                ["aat_wrest", arg_list[0], arg_list[1]] + arg_list[2],
                stdout=subprocess.PIPE,
            ) as wf_metadata:
                wf_json = wf_metadata.communicate()[0]
                logger.info(wf_json.decode())
                if "ERROR" not in wf_json.decode():
                    merged_args = {
                        **json.loads(wf_json),
                        **wf_request.argument,
                        "relative_path": str(tempdir).split("/")[-1],
                        "root_directory": self.processing_settings.rootDirectory,
                        "ramInGb": self.processing_settings.ramInGb,
                    }
                    wf_request.argument = merged_args
                    if arg_list[0] in ["--product", "--carta"]:
                        self.determine_multiple_productfetch(wf_request)

                    return wf_definition.render_templates(wf_request.argument)
                else:
                    logger.error(wf_json.decode())
                    logger.info("SENDING WORKFLOW FAIL MESSAGE!")
                    failed_msg = WorkflowMessageArchitect(request=wf_request).compose_message("workflow_failed")
                    self.messenger.send_message(**failed_msg)
                    return wf_request

    def _determine_usable_files(self, request: WorkflowRequest, templated_files: List[AbstractFile]) -> List:
        # Override templates if user supplied file has same name and is a valid input file
        usable_templates = []
        usable_files = []
        forbidden_templates = self._get_forbidden_templates_list(request)
        rq_names = []

        # create list of filenames of provided request files for easy sorting
        for file in request.files:
            rq_names.append(file.filename)

        for template in templated_files:
            # check if there is not a newer version supplied, add the template
            if template.filename not in rq_names:
                usable_templates.append(template)

            # check if a newer version was supplied and is not in the list of restricted
            # access templates
            if template.filename in rq_names and template.filename in forbidden_templates:
                # the user supplied a file with the same name as the template, BUT the file is on the forbidden list
                # use the template, not the user's file
                usable_templates.append(template)
                logger.info(
                    f"Cannot append user file {filter(lambda f: f.filename == template.filename, request.files)} to "
                    f"workflow request #{request.workflow_request_id}. "
                    f"Users are not allowed to override required system templates."
                )

        # check that supplied files are not in conflict with the sorted templates
        for file in request.files:
            if file.filename not in usable_templates:
                usable_files.append(file)

        for file in usable_templates:
            self.info.save_file(request=request, filename=file.filename, content=file.content)

        return usable_templates + usable_files

    @staticmethod
    def _prepare_files_for_condor(files: List[WorkflowRequestFile], temp_folder: Path):
        """
        Place the files for Condor into a new temp directory and returns the directory.

        :param files:  a dictionary of filename -> content
        :return:       a Path
        """

        # 1. spool each of the temp files to tmp directory
        for file in files:
            path = temp_folder / file.filename
            logger.info("Writing %s", path)
            path.write_bytes(file.content)

        # 2. make any scripts in there executable
        for file in temp_folder.glob("*.sh"):
            logger.info("Making %s executable", file)
            file.chmod(file.stat().st_mode | stat.S_IEXEC)

    @staticmethod
    def _check_for_old_condor_files(dir_path: Path):
        """
        It is possible that a follow-on workflow will use an existing, previously used, directory, and this directory
        might contain previously used condor submit files and their associated output. Move these aside so they don't
        interfere with the new workflow execution
        Example: VLASS Image Caching post SECI or Quicklook

        :param dir_path: path of the directory that will be submitted to condor
        :return:
        """
        extensions = ["*.condor", "*.dag", "*.log", "*.sh", "*.err.txt", "*.out.txt", "*.err", "*.out"]

        old_files = itertools.chain.from_iterable(dir_path.glob(pattern) for pattern in extensions)
        found = list(old_files)
        if len(found) > 0:
            logger.info("Shifting old condor files aside...")
            aside = dir_path / "old_condor"
            aside.mkdir(0o777, parents=False, exist_ok=True)
            for file in found:
                shutil.move(str(file), str(aside))
            logger.info("Finished shifting old condor files.")

    def _execute_prepared(self, folder: Path) -> [Path, bool]:
        """
        Execute HTCondor using the named folder as the source of the files.

        :param folder:  the path to the folder to execute
        :return:        the path to the log file
        """
        logger.info("executing on folder %s", folder)

        # some file in here should end in .dag; that file is our dagman input
        dag_files = list(folder.glob("*.dag"))
        if dag_files:
            logger.info("dagman file %s exists.", dag_files[0])
            return [self._submit_dag(folder, dag_files[0]), True]
        else:
            job_file = list(folder.glob("*.condor"))[0]
            logger.info("condor file %s exists.", job_file)
            return [self._submit_job(folder, job_file), False]

    def _submit_job(self, folder: Path, job_file: Path) -> Path:
        """
        Submit job file to HTCondor

        :param folder: Folder to execute workflow in
        :param job_file: Path to job submit file
        :return: PAth to workflow log file
        """
        # ensure the log file exists
        logfile = self._get_job_logfile_name(job_file)
        logger.info("log file %s exists.", logfile)
        logfile.touch()

        # submit
        logger.info("submitting job to condor...")
        subprocess.run(
            ["condor_submit", str(job_file)],
            cwd=str(folder.absolute()),
            preexec_fn=self._switch_to_submituser(),
        )

        # return the logfile
        return logfile

    def _submit_dag(self, folder: Path, dag_file: Path) -> Path:
        """
        Submit DAG file to HTCondor

        :param folder: Folder to execute workflow in
        :param dag_file: Path to DAG submit file
        :return: Path to workflow log file
        """
        # ensure the log file exists
        logfile = self._get_dag_logfile_name(dag_file)
        logger.info("log file %s exists.", logfile)
        logfile.touch()

        # submit
        logger.info("submitting DAG to condor...")
        subprocess.run(
            ["condor_submit_dag", "-f", str(dag_file)],
            cwd=str(folder.absolute()),
            preexec_fn=self._switch_to_submituser(),
        )

        # return the logfile
        return logfile

    @staticmethod
    def _switch_to_submituser():
        """
        Helper function that is meant to be called as a `preexec_fn` for `subprocess.run`;
        sets gid and uid of subprocess to those of the submituser so that jobs can be
        properly submitted
        """
        # only valid submituser currently is vlapipe
        submituser_gid = 6000
        submituser_uid = 6000

        os.setgid(submituser_gid)
        os.setuid(submituser_uid)

    @staticmethod
    def _get_job_logfile_name(jobfile_name: Path) -> Path:
        """
        Read HTCondor job file and get the log file name, if it exists

        :param jobfile_name: Job file name
        :return: Log file name or "{job-name}.log"
        """
        with open(jobfile_name, "rt") as jobfile:
            for line in jobfile:
                if line.startswith("log = "):
                    _, logfile_name = line.split(" = ")
                    return Path(logfile_name.strip())
        return Path(f"{jobfile_name.stem}.log")

    @staticmethod
    def _get_dag_logfile_name(dag_file: Path) -> Path:
        """
        Return path to DAG log file, which will always be {dag-file-name}.dagman.log

        :param dag_file: Path to workflow DAG file
        :return: Path to workflow DAG log
        """
        return Path(str(dag_file) + ".dagman.log")

    @staticmethod
    def _get_forbidden_templates_list(wf_request: WorkflowRequest) -> List[str]:
        """
        Returns list of templates users are not allowed to override for a specified workflow

        :param wf_request: the workflow request to determine forbidden templates for
        :return: list of forbidden templates
        """
        forbidden = []

        wf_templates = wf_request.workflow.templates

        for template in wf_templates:
            filename = template.filename
            if filename.startswith(wf_request.workflow_name) and (
                filename.endswith(".sh") or filename.endswith(".condor") or filename.endswith(".dag")
            ):
                forbidden.append(filename)

        return forbidden

    def announce_qa(self, workflow_request_id: int, msg_type: str, capability_version: int):
        """
        Announce this QA ready/pass/fail event
        :param workflow_request_id: the workflow request ID
        :param msg_type: ready, pass, or fail
        :param capability_version: version that requires qa
        :return:
        """
        logger.info(f"ANNOUNCING {msg_type.upper()} for workflow request #{workflow_request_id}!")
        wf_request = self.info.lookup_workflow_request(workflow_request_id)

        qa_event_msg = WorkflowMessageArchitect(request=wf_request, ui_info=capability_version).compose_message(
            msg_type
        )
        self.messenger.send_message(**qa_event_msg)

        return qa_event_msg

    def send_forced_fail(self, wf_request_id: int):
        """
        manually sending a fail message

        :param wf_request_id: id of request to fail
        :return:
        """
        wf_request = self.info.lookup_workflow_request(wf_request_id)
        fail_msg = WorkflowMessageArchitect(request=wf_request).compose_message("workflow_failed")
        self.messenger.send_message(**fail_msg)


class WorkflowMessageHandler:
    """Handles workflow messages. No, really!"""

    ITERATION_LIMIT = 3

    def __init__(self, info: WorkflowInfoIF):
        self.info = info
        self.message_router = Router("workflow")
        self.message_router.register(self)
        self.messenger = MessageSender("workflow")
        self.archive_messenger = MessageSender("archive")
        self.url = CapoConfig().settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl

    @on_message(service="workflow", type="workflow-complete")
    def propagate_delivery(self, **message: Dict):
        """
        Deliver this message.

        :param message: message to be delivered
        :return:
        """
        logger.debug("Received workflow complete message. Sending delivery message.")
        # look up the workflow request to get the path
        wf_request = self.info.lookup_workflow_request(message["subject"]["workflow_request_id"])
        controller = self.info.lookup_controller(wf_request.controller)

        # find the path to the delivery.json file
        delivery_file = pathlib.Path(wf_request.results_dir) / "delivery.json"

        # does it exist? if so, we need to send out a new event with its contents
        if delivery_file.exists():
            logger.debug("Found delivery.json file, sending delivery notification")
            with open(delivery_file, "r") as delivery_fo:
                delivery_info = json.load(delivery_fo)

                if controller.control_system == "WS":
                    # construct the WS message
                    msg = WorkflowMessageArchitect(
                        previous_info=message["subject"], delivery_info=delivery_info
                    ).compose_message("delivery")

                    # send the message
                    self.messenger.send_message(**msg)
                elif controller.control_system == "AAT":
                    # construct the AAT message
                    routing_key = f"ws-workflow.download.{wf_request.workflow_request_id}"
                    msg = ArchiveMessageArchitect(
                        routing_key=routing_key, request=wf_request, delivery_info=delivery_info
                    ).compose_message("download_complete")
                    self.archive_messenger.send_message(**msg)
        else:
            logger.debug("No delivery.json found")

    @on_message(service="workflow", type="workflow-complete")
    def send_ingestion_qa_complete_message(self, **message: Dict):
        """
        Send AMQP message to notify the system that ingestion has succeeded

        :param message: workflow-complete message body
        """
        subject = message["subject"]
        wf_req_id = subject["workflow_request_id"]
        wf_request = self.info.lookup_workflow_request(wf_req_id)

        if "ingest" in wf_request.workflow_name and "obs" not in wf_request.workflow_name:
            if "seci" not in wf_request.workflow_name:
                subject["execution_wf_id"] = wf_request.argument["parent_wf_request_id"]
                logger.debug("Sending ingestion complete message!")

                ingestion_complete_msg = WorkflowMessageArchitect(previous_info=subject).compose_message(
                    "ingestion_complete"
                )
                self.messenger.send_message(**ingestion_complete_msg)
            else:
                # Write SECI ingestion results to directory for pre-cleanup result check
                logger.debug("Writing ingestion complete file for VLASS!")
                self.write_vlass_result_file(wf_request, message)

        if "qa" in wf_request.workflow_name:
            subject["execution_wf_id"] = wf_request.argument["parent_wf_request_id"]
            logger.debug("Sending QA Fail Complete message!")
            qa_fail_complete_msg = WorkflowMessageArchitect(previous_info=subject).compose_message("qa_complete")
            self.messenger.send_message(**qa_fail_complete_msg)

    @on_message(service="workflow", type="workflow-failed")
    def send_ingestion_failed_message(self, **message: Dict):
        """
        Send AMQP message to notify the system that ingestion has failed

        :param message: workflow-failed message body
        """
        subject = message["subject"]
        wf_req_id = subject["workflow_request_id"]
        wf_request = self.info.lookup_workflow_request(wf_req_id)

        if "ingest" in wf_request.workflow_name and "obs" not in wf_request.workflow_name:
            if "seci" not in wf_request.workflow_name:
                subject["execution_wf_id"] = wf_request.argument["parent_wf_request_id"]
                logger.debug("Sending ingestion failed message!")

                ingestion_failed_msg = WorkflowMessageArchitect(previous_info=subject).compose_message(
                    "ingestion_failed"
                )
                self.messenger.send_message(**ingestion_failed_msg)
            else:
                # Write SECI ingestion results to directory for pre-cleanup result check
                logger.debug("Writing ingestion failed to file for VLASS!")
                self.write_vlass_result_file(wf_request, message)

    @on_message(service="workflow")
    def on_workflow_event(self, **message: Dict):
        """
        Handle the workflow event indicated by this message.

        :param message:
        :return:
        """
        logger.debug(f"received workflow message: {message}")
        msg_subject = message["subject"]
        if msg_subject["type"] == "WorkflowRequest":
            request_id = msg_subject["workflow_request_id"]
            request = self.info.lookup_workflow_request(request_id)

            if request:
                self._process_message_for_request(message, request)
            else:
                logger.warning(
                    f"No workflow request found for request id #{request_id}, from message subject {msg_subject}."
                )
        else:
            logger.warning(f"Message {message} does not concern a workflow request. Ignoring.")

    def _process_message_for_request(self, message: Dict, request: WorkflowRequest):
        # update request record with new status
        if message["type"] == "workflow-executing":
            status = WorkflowRequestState.Running.name if self._verify_state_change(request.state) else None

            # on initiation, save the HTC id for this workflow:
            if htcondor_job_id := int(message["condor_metadata"]["condor_job_id"]):
                # Workflow has corresponding condor job ID
                logger.info(f"Workflow request has an HTCondor job ID of {htcondor_job_id}.")
                request.htcondor_job_id = htcondor_job_id
                request.htcondor_iterations = 1
                # update UI
                iterations_msg = WorkflowMessageArchitect(
                    request=request, json_payload={"htcondor_iterations": request.htcondor_iterations}
                ).compose_message("update_wf_metadata")
                self.messenger.send_message(**iterations_msg)

        elif message["type"] == "workflow-updated":
            status = WorkflowRequestState.Running.name if self._verify_state_change(request.state) else None

            # update workflow_progress table
            stage_info = self._update_workflow_progress(request, message)
            if isinstance(stage_info, Dict):
                # message relevant control system
                if any(c.control_system == request.controller for c in self.info.all_external_controllers()):
                    tack_on = {**message, **stage_info}
                    self.send_external_event("update", **tack_on)

        elif message["type"] == "workflow-failed":
            request = self.info.refresh_request(request)
            status = WorkflowRequestState.Error.name
            self._post_workflow_cleanup(message, request, "failed")

        elif message["type"] == "workflow-aborted":
            request = self.info.refresh_request(request)
            status = WorkflowRequestState.Failed.name
            self._post_workflow_cleanup(message, request, "failed")

        elif message["type"] == "workflow-continuing":
            status = WorkflowRequestState.Running.name if self._verify_state_change(request.state) else None

        elif message["type"] in ("update-wf-metadata", "ingestion-failed"):
            # no action, keep existing state:
            return

        elif message["type"] == "workflow-paused":
            status = WorkflowRequestState.Paused.name

            user_message = (
                f"HTCondor has paused request #{request.workflow_request_id}. "
                f"Please check your input files for errors."
            )
            logger.warning(user_message)

            # update UI
            paused_msg = WorkflowMessageArchitect(
                request=request,
                json_payload={
                    "user_warning_message": user_message,
                },
            ).compose_message("update_wf_metadata")
            self.messenger.send_message(**paused_msg)

        elif message["type"] == "workflow-restarted":
            request.htcondor_iterations = request.htcondor_iterations + 1

            if request.htcondor_iterations > self.ITERATION_LIMIT:
                status = WorkflowRequestState.Error
                user_message = (
                    f"Found possible infinite rescheduling loop in "
                    f"request #{request.workflow_request_id}, HTCondor iteration "
                    f"count is {request.htcondor_iterations}. Aborting...."
                )
                logger.error(user_message)
                url = self.url + f"/workflows/requests/{request.workflow_request_id}/abort"
                requests.post(url)
            else:
                status = WorkflowRequestState.Running.name
                user_message = (
                    f"Workflow request #{request.workflow_request_id} rescheduled! "
                    f"HTCondor iteration count is {request.htcondor_iterations}. Request will "
                    f"abort if count reaches {self.ITERATION_LIMIT + 1}."
                )
                logger.warning(user_message)

            # update UI
            iterations_msg = WorkflowMessageArchitect(
                request=request,
                json_payload={
                    "htcondor_iterations": request.htcondor_iterations,
                    "user_warning_message": user_message,
                },
            ).compose_message("update_wf_metadata")
            self.messenger.send_message(**iterations_msg)

        elif message["type"] == "workflow-complete":
            request = self.info.refresh_request(request)
            status = WorkflowRequestState.Complete.name
            self._post_workflow_cleanup(message, request, "complete")

        elif self._complete_no_cleanup(message):
            status = WorkflowRequestState.Complete.name
        else:
            status = "Unknown"

        logger.info(
            "Setting state on workflow request %s to %s...",
            request.workflow_request_id,
            status,
        )

        # update request record with new status
        try:
            if status is not None:
                request.update_status(status)
            self.info.save_request(request)
            transaction.commit()
        except Exception as exc:
            transaction.abort()
            logger.error(
                f"Failed to update status on workflow request {request.workflow_request_id} to {status}: {exc}"
            )

    @staticmethod
    def _verify_state_change(current_state: str) -> bool:
        """
        Sometimes messages come in out of order. Make sure Complete/Failed/Error states don't change once set.

        :param current_state: Verify this request in not in a final state before persisting a state change
        :return: boolean
        """
        final_list = [
            WorkflowRequestState.Error.name,
            WorkflowRequestState.Complete.name,
            WorkflowRequestState.Failed.name,
        ]

        return current_state not in final_list

    @staticmethod
    def _complete_no_cleanup(message: Dict):
        no_cleanup_messages = ["delivery", "ingestion-complete"]

        if message["type"] in no_cleanup_messages:
            return True
        elif "qa" in message["type"]:
            return True
        else:
            return False

    def _post_workflow_cleanup(self, message: Dict, request: WorkflowRequest, event_type: str):
        if any(c.control_system == request.controller for c in self.info.all_external_controllers()):
            self.send_external_event(event_type, **message)

        self.clean_remote_workflow(request)
        self.clean_workflow(request)

    def _update_workflow_progress(self, request: WorkflowRequest, message: dict) -> Union[Dict, bool]:
        """
        Update current stage of running workflow to workflow_progress table

        :param message: message to update from
        :return:
        """
        start = None
        end = None

        condor_msg = message["condor_metadata"]
        timestamp = condor_msg["timestamp"]
        log = condor_msg["log"]

        parse_log = log.split("Stage", 1)[1].split("has", 1)

        stage = parse_log[0].strip()
        msg_type = parse_log[1].split("at", 1)[0].strip()

        if msg_type == "Started":
            start = timestamp
        else:
            end = timestamp

        stage_info = self.info.lookup_workflow_stage(request.workflow_request_id, stage)

        if stage_info and stage_info.end is not None:
            # This entry already exists in progress table, message is a duplicate. Ignore
            return False

        self.info.save_progress(request, stage, start, end)
        return {
            "stage_info": {
                "stage": stage,
                "start": start,
                "stop": end,
            }
        }

    def send_external_event(self, event_type: str, **message: Dict):
        """
        Create and send AMQP event to non-WS messaging service

        :param message: message to send
        :param event_type: type of event to send (complete, fail, update)
        :return:
        """
        subject = message["subject"]
        wf_id = subject["workflow_request_id"]
        wf_request = self.info.lookup_workflow_request(wf_id)
        controller = self.info.lookup_controller(wf_request.controller)
        wf_key = None
        stage_info = None

        if event_type == "update":
            stage_info = message["stage_info"]

        logger.debug(
            f"Sending {wf_request.workflow_name} {event_type} message to "
            f"{controller.control_system} for request #{wf_id}!"
        )

        if controller.control_system == "AAT":
            # download_complete will be sent from propagate_delivery method
            if wf_request.workflow_name == "download" and event_type == "complete":
                return

            if wf_request.workflow_name == "carta":
                wf_key = "carta-instance-ready"
                if event_type == "complete":
                    # carta exist by killing its processes which condor reads as an error.
                    # A 'normal' exit 0 for carta is actually a failure and the carta_complete message doesn't exist.
                    # fix message type
                    event_type = "failed"
            else:
                wf_key = wf_request.workflow_name

        if controller.control_system == "VLASS":
            wf_key = "caching" if "caching" in wf_request.workflow_name else wf_request.workflow_name.split("_", 1)[1]
            # fix for cubes
            if "coarse" == wf_key:
                wf_key = wf_key + "_cube"

        routing_key = f"{controller.routing_key_prefix}{wf_key}.{wf_id}"
        msg_type = f"{wf_request.workflow_name}_{event_type}"
        if routing_key is not None:
            if controller.message_handler == "archive":
                if event_type == "update":
                    architect = ArchiveMessageArchitect(
                        routing_key=routing_key, request=wf_request, stage_info=stage_info
                    )
                else:
                    architect = ArchiveMessageArchitect(routing_key=routing_key, request=wf_request)
                msg = architect.compose_message(msg_type)
                self.archive_messenger.send_message(**msg)
            else:
                logger.warning(
                    f"Unable to determine messenger service for request #{wf_id}"
                    f" with control system {controller.control_system}!"
                )
        else:
            logger.warning(f"Unable to determine routing key for request #{wf_id}! Skipping message!")

    def clean_remote_workflow(self, request: WorkflowRequest):
        injector = CapoInjector(self.info, request)

        if injector.is_remote_workflow():
            logger.debug("Cleaning remote workflow")
            result = injector.clear_subspace()
            if result is False:
                # the processing directory somehow disappeared, mark as cleaned to avoid further errors
                request.cleaned = True

    @staticmethod
    def clean_workflow(request: WorkflowRequest):
        if request.htcondor_job_id is not None:
            # Request has a set condor_job_id; unset it
            logger.info(f"Workflow #{request.workflow_request_id} finished; unsetting HTCondor ID")
            request.htcondor_job_id = None

        if request.controller == "VLASS" and "vlass_" in request.results_dir:
            # request is initiated and controlled by VLASS Manager cleanup system.
            # Mark request as cleaned to bypass WS Annihilator and prevent clogging the logs
            request.cleaned = True

    @staticmethod
    def write_vlass_result_file(wf_request: WorkflowRequest, message: dict):
        file_path = wf_request.results_dir + "/ingest-result.json"
        result_file = open(file_path, "x")
        result_file.write(json.dumps(message))
        result_file.close()
